@@ -1,6 +1,7 @@
 # Orca
 #
 # Copyright 2009 Sun Microsystems Inc.
+# Copyright 2015-2016 Igalia, S.L.
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -22,22 +23,23 @@
 __id__        = "$Id:$"
 __version__   = "$Revision:$"
 __date__      = "$Date:$"
-__copyright__ = "Copyright (c) 2009 Sun Microsystems Inc."
+__copyright__ = "Copyright (c) 2009 Sun Microsystems Inc." \
+                "Copyright (c) 2015-2016 Igalia, S.L."
 __license__   = "LGPL"
 
+import collections
+import pyatspi
 import sys
 import time
 import traceback
-
-import pyatspi
+from gi.repository import Atspi, Atk
 
 from . import braille
 from . import debug
 from . import messages
 from . import object_properties
 from . import settings
-
-import collections
+from . import settings_manager
 
 def _formatExceptionInfo(maxTBlevel=5):
     cla, exc, trbk = sys.exc_info()
@@ -58,6 +60,8 @@ def _formatExceptionInfo(maxTBlevel=5):
 #
 METHOD_PREFIX = "_generate"
 
+_settingsManager = settings_manager.getManager()
+
 class Generator:
     """Takes accessible objects and generates a presentation for those
     objects.  See the generate method, which is the primary entry
@@ -71,6 +75,7 @@ class Generator:
 
         self._mode = mode
         self._script = script
+        self._activeProgressBars = {}
         self._methodsDict = {}
         for method in \
             [z for z in [getattr(self, y).__get__(self, self.__class__) for y in [x for x in dir(self) if x.startswith(METHOD_PREFIX)]] if isinstance(z, collections.Callable)]:
@@ -96,7 +101,7 @@ class Generator:
         # for the results of all the legal method names.
         #
         globalsDict = {}
-        for key in list(self._methodsDict.keys()):
+        for key in self._methodsDict.keys():
             globalsDict[key] = []
         self._addGlobals(globalsDict)
 
@@ -123,16 +128,9 @@ class Generator:
                             arg = arg.replace("' is not defined", "")
                             if arg not in self._methodsDict:
                                 debug.printException(debug.LEVEL_SEVERE)
-                                debug.println(
-                                    debug.LEVEL_SEVERE,
-                                    "Unable to find function for '%s'\n" % arg)
                             globalsDict[arg] = []
                         except:
                             debug.printException(debug.LEVEL_SEVERE)
-                            debug.println(
-                                debug.LEVEL_SEVERE,
-                                "While processing '%s' '%s' '%s' '%s'" \
-                                % (roleKey, key, evalString, globalsDict))
                             break
 
     def _overrideRole(self, newRole, args):
@@ -188,8 +186,8 @@ class Generator:
         try:
             globalsDict['role'] = args.get('role', obj.getRole())
         except:
-            msg = 'Cannot generate presentation for: %s. Aborting' % obj
-            debug.println(debug.LEVEL_FINEST, msg)
+            msg = 'ERROR: Cannot generate presentation for: %s. Aborting' % obj
+            debug.println(debug.LEVEL_INFO, msg, True)
             return result
         try:
             # We sometimes want to override the role.  We'll keep the
@@ -226,17 +224,8 @@ class Generator:
             else:
                 firstTimeCalled = False
 
-            details = debug.getAccessibleDetails(debug.LEVEL_ALL, obj)
-            duration = "%.4f" % (time.time() - startTime)
-            debug.println(debug.LEVEL_ALL, "\nPREPARATION TIME: %s" % duration)
-            debug.println(
-                debug.LEVEL_ALL,
-                "generate %s for %s %s\n(args=%s)\nusing '%s'" \
-                % (self._mode,
-                   args['formatType'], 
-                   details,
-                   repr(args),
-                   formatting))
+            msg = '%s GENERATOR: Starting generation for %s' % (self._mode.upper(), obj)
+            debug.println(debug.LEVEL_INFO, msg, True)
 
             assert(formatting)
             while True:
@@ -252,24 +241,25 @@ class Generator:
                     arg = arg.replace("' is not defined", "")
                     if arg not in self._methodsDict:
                         debug.printException(debug.LEVEL_SEVERE)
-                        debug.println(
-                            debug.LEVEL_SEVERE,
-                            "Unable to find function for '%s'\n" % arg)
                         break
                     globalsDict[arg] = self._methodsDict[arg](obj, **args)
                     duration = "%.4f" % (time.time() - currentTime)
                     debug.println(debug.LEVEL_ALL,
-                                  "GENERATION  TIME: %s  ---->  %s=%s" \
+                                  "           GENERATION TIME: %s  ---->  %s=%s" \
                                   % (duration, arg, repr(globalsDict[arg])))
+
         except:
             debug.printException(debug.LEVEL_SEVERE)
             result = []
 
         duration = "%.4f" % (time.time() - startTime)
-        debug.println(debug.LEVEL_ALL, "COMPLETION  TIME: %s" % duration)
-        debug.println(debug.LEVEL_ALL, "generate %s results:" % self._mode)
+        debug.println(debug.LEVEL_ALL, "           COMPLETION TIME: %s" % duration)
+        debug.println(debug.LEVEL_ALL, "%s GENERATOR: Results:" % self._mode.upper(), True)
         for element in result:
-            debug.println(debug.LEVEL_ALL, "  %s" % element)
+            debug.println(debug.LEVEL_ALL, "           %s" % element)
+
+        if args.get('isProgressBarUpdate') and result:
+            self.setProgressBarUpdateTimeAndValue(obj)
 
         return result
 
@@ -326,11 +316,12 @@ class Generator:
                 return result
             if description:
                 result.append(description)
+                self._script.pointOfReference['usedDescriptionForName'] = True
             else:
                 link = None
                 if obj.getRole() == pyatspi.ROLE_LINK:
                     link = obj
-                elif obj.parent.getRole() == pyatspi.ROLE_LINK:
+                elif obj.parent and obj.parent.getRole() == pyatspi.ROLE_LINK:
                     link = obj.parent
                 if link:
                     basename = self._script.utilities.linkBasename(link)
@@ -390,6 +381,18 @@ class Generator:
         represent the description of the object, if that description
         is different from that of the name and label.
         """
+
+        if self._script.pointOfReference.get('usedDescriptionForName'):
+            return []
+
+        role = args.get('role', obj.getRole())
+
+        # Unity Panel Service menubar items are labels which claim focus and
+        # have an accessible description of the text + underscore symbol used
+        # to create the mnemonic. We'll work around that here for now.
+        if role == pyatspi.ROLE_LABEL:
+            return []
+
         result = []
         if obj.description:
             label = self._script.utilities.displayedLabel(obj) or ""
@@ -610,6 +613,22 @@ class Generator:
                 result.append(indicators[0])
         return result
 
+    def _generateMultiselectableState(self, obj, **args):
+        """Returns an array of strings (and possibly voice and audio
+        specifications) that represent the multiselectable state of
+        the object.  This is typically for list boxes. If the object
+        is not multiselectable, an empty array will be returned.
+        """
+
+        result = []
+        if not args.get('mode', None):
+            args['mode'] = self._mode
+        args['stringType'] = 'multiselect'
+        if obj.getState().contains(pyatspi.STATE_MULTISELECTABLE) \
+           and obj.childCount:
+            result.append(self._script.formatting.getString(**args))
+        return result
+
     #####################################################################
     #                                                                   #
     # Table interface information                                       #
@@ -630,7 +649,7 @@ class Generator:
         if not text:
             return result
 
-        roleString =  self.getLocalizedRoleName(obj, pyatspi.ROLE_ROW_HEADER)
+        roleString =  self.getLocalizedRoleName(obj, role=pyatspi.ROLE_ROW_HEADER)
         if args.get('mode') == 'speech':
             if settings.speechVerbosityLevel == settings.VERBOSITY_LEVEL_VERBOSE \
                and not args.get('formatType') in ['basicWhereAmI', 'detailedWhereAmI']:
@@ -656,7 +675,7 @@ class Generator:
         if not text:
             return result
 
-        roleString =  self.getLocalizedRoleName(obj, pyatspi.ROLE_COLUMN_HEADER)
+        roleString =  self.getLocalizedRoleName(obj, role=pyatspi.ROLE_COLUMN_HEADER)
         if args.get('mode') == 'speech':
             if settings.speechVerbosityLevel == settings.VERBOSITY_LEVEL_VERBOSE \
                and not args.get('formatType') in ['basicWhereAmI', 'detailedWhereAmI']:
@@ -794,7 +813,8 @@ class Generator:
         except:
             parentTable = None
         isDetailedWhereAmI = args.get('formatType', None) == 'detailedWhereAmI'
-        if (settings.readTableCellRow or isDetailedWhereAmI) and parentTable \
+        readFullRow = self._script.utilities.shouldReadFullRow(obj)
+        if (readFullRow or isDetailedWhereAmI) and parentTable \
            and (not self._script.utilities.isLayoutOnly(obj.parent)):
             parent = obj.parent
             index = self._script.utilities.cellIndex(obj)
@@ -861,8 +881,7 @@ class Generator:
             return []
 
         substring = self._script.utilities.substring(obj, start, end)
-        if substring and substring.strip() != obj.name \
-           and not self._script.EMBEDDED_OBJECT_CHARACTER in substring:
+        if substring and not self._script.EMBEDDED_OBJECT_CHARACTER in substring:
             return [substring]
 
         return []
@@ -1047,3 +1066,179 @@ class Generator:
 
     def _generatePageSummary(self, obj, **args):
         return []
+
+    def _generateProgressBarIndex(self, obj, **args):
+        return []
+
+    def _generateProgressBarValue(self, obj, **args):
+        return []
+
+    def _getProgressBarUpdateInterval(self):
+        return int(_settingsManager.getSetting('progressBarUpdateInterval'))
+
+    def _shouldPresentProgressBarUpdate(self, obj, **args):
+        percent = self._script.utilities.getValueAsPercent(obj)
+        lastTime, lastValue = self.getProgressBarUpdateTimeAndValue(obj, type=self)
+        if percent == lastValue:
+            return False
+
+        if percent == 100:
+            return True
+
+        interval = int(time.time() - lastTime)
+        return interval >= self._getProgressBarUpdateInterval()
+
+    def _cleanUpCachedProgressBars(self):
+        isValid = lambda x: not (self._script.utilities.isZombie(x) or self._script.utilities.isDead(x))
+        bars = list(filter(isValid, self._activeProgressBars))
+        self._activeProgressBars = {x:self._activeProgressBars.get(x) for x in bars}
+
+    def _getMostRecentProgressBarUpdate(self):
+        self._cleanUpCachedProgressBars()
+        if not self._activeProgressBars.values():
+            return None, 0.0, None
+
+        sortedValues = sorted(self._activeProgressBars.values(), key=lambda x: x[0])
+        prevTime, prevValue = sortedValues[-1]
+        return list(self._activeProgressBars.keys())[-1], prevTime, prevValue
+
+    def getProgressBarNumberAndCount(self, obj):
+        self._cleanUpCachedProgressBars()
+        if not obj in self._activeProgressBars:
+            self._activeProgressBars[obj] = 0.0, None
+
+        thisValue = self.getProgressBarUpdateTimeAndValue(obj)
+        index = list(self._activeProgressBars.values()).index(thisValue)
+        return index + 1, len(self._activeProgressBars)
+
+    def getProgressBarUpdateTimeAndValue(self, obj, **args):
+        if not obj in self._activeProgressBars:
+            self._activeProgressBars[obj] = 0.0, None
+
+        return self._activeProgressBars.get(obj)
+
+    def setProgressBarUpdateTimeAndValue(self, obj, lastTime=None, lastValue=None):
+        lastTime = lastTime or time.time()
+        lastValue = lastValue or self._script.utilities.getValueAsPercent(obj)
+        self._activeProgressBars[obj] = lastTime, lastValue
+
+    def _getAlternativeRole(self, obj, **args):
+        if self._script.utilities.isMath(obj):
+            if self._script.utilities.isMathFraction(obj):
+                return 'ROLE_MATH_FRACTION'
+            if self._script.utilities.isMathRoot(obj):
+                return 'ROLE_MATH_ROOT'
+            if self._script.utilities.isMathSubOrSuperScript(obj):
+                return 'ROLE_MATH_SCRIPT_SUBSUPER'
+            if self._script.utilities.isMathUnderOrOverScript(obj):
+                return 'ROLE_MATH_SCRIPT_UNDEROVER'
+            if self._script.utilities.isMathMultiScript(obj):
+                return 'ROLE_MATH_MULTISCRIPT'
+            if self._script.utilities.isMathEnclose(obj):
+                return 'ROLE_MATH_ENCLOSED'
+            if self._script.utilities.isMathFenced(obj):
+                return 'ROLE_MATH_FENCED'
+            if self._script.utilities.isMathTable(obj):
+                return 'ROLE_MATH_TABLE'
+            if self._script.utilities.isMathTableRow(obj):
+                return 'ROLE_MATH_TABLE_ROW'
+        if self._script.utilities.isStatic(obj):
+            return 'ROLE_STATIC'
+        if self._script.utilities.isFocusableLabel(obj):
+            return pyatspi.ROLE_LIST_ITEM
+
+        return args.get('role', obj.getRole())
+
+    def getLocalizedRoleName(self, obj, **args):
+        role = args.get('role', obj.getRole())
+
+        if "Value" in pyatspi.listInterfaces(obj):
+            state = obj.getState()
+            isVertical = state.contains(pyatspi.STATE_VERTICAL)
+            isHorizontal = state.contains(pyatspi.STATE_HORIZONTAL)
+            isFocused = state.contains(pyatspi.STATE_FOCUSED) \
+                        or args.get('alreadyFocused', False)
+
+            if role == pyatspi.ROLE_SLIDER:
+                if isHorizontal:
+                    return object_properties.ROLE_SLIDER_HORIZONTAL
+                if isVertical:
+                    return object_properties.ROLE_SLIDER_VERTICAL
+            elif role == pyatspi.ROLE_SCROLL_BAR:
+                if isHorizontal:
+                    return object_properties.ROLE_SCROLL_BAR_HORIZONTAL
+                if isVertical:
+                    return object_properties.ROLE_SCROLL_BAR_VERTICAL
+            elif role == pyatspi.ROLE_SEPARATOR:
+                if isHorizontal:
+                    return object_properties.ROLE_SPLITTER_HORIZONTAL
+                if isVertical:
+                    return object_properties.ROLE_SPLITTER_VERTICAL
+            elif role == pyatspi.ROLE_SPLIT_PANE and isFocused:
+                # The splitter has the opposite orientation of the split pane.
+                if isHorizontal:
+                    return object_properties.ROLE_SPLITTER_VERTICAL
+                if isVertical:
+                    return object_properties.ROLE_SPLITTER_HORIZONTAL
+
+        if self._script.utilities.isLandmark(obj):
+            if self._script.utilities.isLandmarkBanner(obj):
+                return object_properties.ROLE_LANDMARK_BANNER
+            if self._script.utilities.isLandmarkComplementary(obj):
+                return object_properties.ROLE_LANDMARK_COMPLEMENTARY
+            if self._script.utilities.isLandmarkContentInfo(obj):
+                return object_properties.ROLE_LANDMARK_CONTENTINFO
+            if self._script.utilities.isLandmarkMain(obj):
+                return object_properties.ROLE_LANDMARK_MAIN
+            if self._script.utilities.isLandmarkNavigation(obj):
+                return object_properties.ROLE_LANDMARK_NAVIGATION
+            if self._script.utilities.isLandmarkRegion(obj):
+                return object_properties.ROLE_LANDMARK_REGION
+            if self._script.utilities.isLandmarkSearch(obj):
+                return object_properties.ROLE_LANDMARK_SEARCH
+            if self._script.utilities.isLandmarkForm(obj):
+                role = pyatspi.ROLE_FORM
+
+        if not isinstance(role, (pyatspi.Role, Atspi.Role)):
+            try:
+                return obj.getLocalizedRoleName()
+            except:
+                return ''
+
+        nonlocalized = Atspi.role_get_name(role)
+        atkRole = Atk.role_for_name(nonlocalized)
+        if atkRole == Atk.Role.INVALID and role == pyatspi.ROLE_STATUS_BAR:
+            atkRole = Atk.role_for_name("statusbar")
+
+        return Atk.role_get_localized_name(atkRole)
+
+    def getStateIndicator(self, obj, **args):
+        role = args.get('role', obj.getRole())
+
+        if role == pyatspi.ROLE_MENU_ITEM:
+            return self._generateMenuItemCheckedState(obj, **args)
+
+        if role in [pyatspi.ROLE_RADIO_BUTTON, pyatspi.ROLE_RADIO_MENU_ITEM]:
+            return self._generateRadioState(obj, **args)
+
+        if role in [pyatspi.ROLE_CHECK_BOX, pyatspi.ROLE_CHECK_MENU_ITEM]:
+            return self._generateCheckedState(obj, **args)
+
+        if role == pyatspi.ROLE_TOGGLE_BUTTON:
+            return self._generateToggleState(obj, **args)
+
+        if role == pyatspi.ROLE_TABLE_CELL:
+            return self._generateCellCheckedState(obj, **args)
+
+        return []
+
+    def getValue(self, obj, **args):
+        role = args.get('role', obj.getRole())
+
+        if role == pyatspi.ROLE_PROGRESS_BAR:
+            return self._generateProgressBarValue(obj, **args)
+
+        if role in [pyatspi.ROLE_SCROLL_BAR, pyatspi.ROLE_SLIDER]:
+            return self._generatePercentage(obj, **args)
+
+        return []
