@@ -29,6 +29,8 @@ __copyright__ = "Copyright (c) 2004-2009 Sun Microsystems Inc." \
                 "Copyright (c) 2012 Igalia, S.L."
 __license__   = "LGPL"
 
+import faulthandler
+import gi
 import importlib
 import os
 import pyatspi
@@ -39,7 +41,7 @@ import sys
 
 try:
     from gi.repository.Gio import Settings
-    a11yAppSettings = Settings('org.gnome.desktop.a11y.applications')
+    a11yAppSettings = Settings(schema_id='org.gnome.desktop.a11y.applications')
 except:
     a11yAppSettings = None
 
@@ -49,10 +51,15 @@ try:
     # for this is to allow "orca --text-setup" to work even if
     # the desktop is not running.
     #
+    gi.require_version("Gtk", "3.0")
     from gi.repository import Gtk
+
+    gi.require_version("Gdk", "3.0")
     from gi.repository import Gdk
+
     # Note: This last import is here due to bgo #673396.
     # See bgo#673397 for the rest of the story.
+    gi.require_version("GdkX11", "3.0")
     from gi.repository.GdkX11 import X11Screen
 except:
     pass
@@ -60,25 +67,23 @@ except:
 from . import braille
 from . import debug
 from . import event_manager
+from . import keybindings
+from . import logger
 from . import messages
 from . import notification_messages
 from . import orca_state
+from . import orca_platform
 from . import script_manager
 from . import settings
 from . import settings_manager
 from . import speech
+from . import sound
 from .input_event import BrailleEvent
-from .input_event import KeyboardEvent
 
 _eventManager = event_manager.getManager()
 _scriptManager = script_manager.getManager()
 _settingsManager = settings_manager.getManager()
-
-try:
-    # If we don't have an active desktop, we will get a RuntimeError.
-    from . import mouse_review
-except RuntimeError:
-    pass
+_logger = logger.getLogger()
 
 def onEnabledChanged(gsetting, key):
     try:
@@ -92,6 +97,9 @@ def onEnabledChanged(gsetting, key):
 def getSettingsManager():
     return _settingsManager
 
+def getLogger():
+    return _logger
+
 EXIT_CODE_HANG = 50
 
 # The user-settings module (see loadUserSettings).
@@ -126,151 +134,51 @@ def setLocusOfFocus(event, obj, notifyScript=True, force=False):
     """
 
     if not force and obj == orca_state.locusOfFocus:
+        msg = "ORCA: Setting locusOfFocus to existing locusOfFocus"
+        debug.println(debug.LEVEL_INFO, msg, True)
         return
 
-    # If this event is not for the currently active script, then just return.
-    #
-    if event and event.source and \
-       event.host_application and orca_state.activeScript:
-        currentApp = orca_state.activeScript.app
-        try:
-            appList = [event.host_application, event.source.getApplication()]
-        except (LookupError, RuntimeError):
-            appList = []
-            debug.println(debug.LEVEL_SEVERE,
-                           "orca.setLocusOfFocus() application Error")
-        if not currentApp in appList:
+    if event and (orca_state.activeScript and not orca_state.activeScript.app):
+        script = _scriptManager.getScript(event.host_application, event.source)
+        _scriptManager.setActiveScript(script, "Setting locusOfFocus")
+
+    oldFocus = orca_state.locusOfFocus
+    try:
+        oldFocus.getRole()
+    except:
+        msg = "ORCA: Old locusOfFocus is null or defunct"
+        debug.println(debug.LEVEL_INFO, msg, True)
+        oldFocus = None
+
+    if not obj:
+        msg = "ORCA: New locusOfFocus is null (being cleared)"
+        debug.println(debug.LEVEL_INFO, msg, True)
+        orca_state.locusOfFocus = None
+        return
+
+    if orca_state.activeScript:
+        if orca_state.activeScript.utilities.isZombie(obj):
+            msg = "ERROR: New locusOfFocus (%s) is zombie. Not updating." % obj
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return
+        if orca_state.activeScript.utilities.isDead(obj):
+            msg = "ERROR: New locusOfFocus (%s) is dead. Not updating." % obj
+            debug.println(debug.LEVEL_INFO, msg, True)
             return
 
-    oldLocusOfFocus = orca_state.locusOfFocus
-    try:
-        # Just to see if we have a valid object.
-        oldLocusOfFocus.getRole()
-    except:
-        # Either it's None or it's an invalid remote object.
-        oldLocusOfFocus = None
-
+    msg = "ORCA: Changing locusOfFocus from %s to %s" % (oldFocus, obj)
+    debug.println(debug.LEVEL_INFO, msg, True)
     orca_state.locusOfFocus = obj
-    try:
-        app = orca_state.locusOfFocus.getApplication()
-    except:
-        orca_state.locusOfFocus = None
-        if event:
-            debug.println(debug.LEVEL_FINE,
-                          "LOCUS OF FOCUS: None event='%s'" % event.type)
-        else:
-            debug.println(debug.LEVEL_FINE,
-                          "LOCUS OF FOCUS: None event=None")
-    else:
-        try:
-            appname = "'" + app.name + "'"
-        except:
-            appname = "None"
-        debug.println(debug.LEVEL_FINE,
-                      "LOCUS OF FOCUS: app=%s name='%s' role='%s'" \
-                      % (appname,
-                         orca_state.locusOfFocus.name,
-                         orca_state.locusOfFocus.getRoleName()))
 
-        if event:
-            debug.println(debug.LEVEL_FINE,
-                          "                event='%s'" % event.type)
-        else:
-            debug.println(debug.LEVEL_FINE,
-                          "                event=None")
-
-    if notifyScript and orca_state.activeScript:
-        orca_state.activeScript.locusOfFocusChanged(
-            event, oldLocusOfFocus, orca_state.locusOfFocus)
-
-########################################################################
-#                                                                      #
-# METHODS FOR PRE-PROCESSING AND MASSAGING KEYBOARD EVENTS.            #
-#                                                                      #
-########################################################################
-
-_orcaModifierPressed = False
-
-def _processKeyboardEvent(event):
-    """The primary key event handler for Orca.  Keeps track of various
-    attributes, such as the lastInputEvent.  Also does key echo as well
-    as any local keybindings before passing the event on to the active
-    script.  This method is called synchronously from the AT-SPI registry
-    and should be performant.  In addition, it must return True if it has
-    consumed the event (and False if not).
-
-    Arguments:
-    - event: an AT-SPI DeviceEvent
-
-    Returns True if the event should be consumed.
-    """
-    global _orcaModifierPressed
-
-    # Weed out duplicate and otherwise bogus events.
-    keyboardEvent = KeyboardEvent(event)
-    debug.println(debug.LEVEL_FINE, keyboardEvent.toString())
-    if keyboardEvent.ignoreDueToTimestamp():
-        debug.println(debug.LEVEL_FINE, "IGNORING EVENT DUE TO TIMESTAMP")
+    if not notifyScript:
         return
 
-    # Figure out what we've got.
-    isOrcaModifier = keyboardEvent.isOrcaModifier()
-    isPressedEvent = keyboardEvent.isPressedKey()
-    if isOrcaModifier:
-        _orcaModifierPressed = isPressedEvent
-    if _orcaModifierPressed:
-        keyboardEvent.modifiers |= settings.ORCA_MODIFIER_MASK
+    if not orca_state.activeScript:
+        msg = "ORCA: Cannot notify active script because there isn't one"
+        debug.println(debug.LEVEL_INFO, msg, True)
+        return
 
-    # Update our state.
-    orca_state.lastInputEventTimestamp = event.timestamp
-    orca_state.lastInputEvent = keyboardEvent
-    if not keyboardEvent.isModifierKey():
-        keyboardEvent.setClickCount()
-        orca_state.lastNonModifierKeyEvent = keyboardEvent
-
-    # Echo it based on what it is and the user's settings.
-    script = orca_state.activeScript
-    if script:
-        if isPressedEvent:
-            script.presentationInterrupt()
-        keyboardEvent.present()
-        if keyboardEvent.isModifierKey() and not isOrcaModifier:
-            return False
- 
-    # Special modes.
-    if not isPressedEvent and keyboardEvent.event_string == "Escape":
-        script.exitLearnMode(keyboardEvent)
-    if orca_state.learnModeEnabled and not keyboardEvent.modifiers:
-        if keyboardEvent.event_string == "F1":
-            orca_state.learnModeEnabled = False
-            return helpForOrca()
-        if keyboardEvent.event_string in ["1", "2"]:
-            orca_state.learnModeEnabled = False
-            return script.listOrcaShortcuts(keyboardEvent)
-    if orca_state.capturingKeys:
-        return False
-    if notification_messages.listNotificationMessagesModeEnabled:
-        return notification_messages.listNotificationMessages(keyboardEvent)
-
-    # See if the event manager wants it (i.e. it is bound to a command.
-    if _eventManager.processKeyboardEvent(keyboardEvent):
-        return True
-
-    # Do any needed xmodmap crap.
-    global _restoreOrcaKeys
-    if not isPressedEvent:
-        if keyboardEvent.event_string in settings.orcaModifierKeys \
-           and orca_state.bypassNextCommand:
-            _restoreXmodmap()
-            _restoreOrcaKeys = True
-        elif _restoreOrcaKeys and not orca_state.bypassNextCommand:
-            _createOrcaXmodmap()
-            _restoreOrcaKeys = False
-    elif not keyboardEvent.isModifierKey():
-        _orcaModifierPressed = False
-        orca_state.bypassNextCommand = False
- 
-    return isOrcaModifier or orca_state.learnModeEnabled
+    orca_state.activeScript.locusOfFocusChanged(event, oldFocus, orca_state.locusOfFocus)
 
 ########################################################################
 #                                                                      #
@@ -312,6 +220,23 @@ def _processBrailleEvent(event):
 #                                                                      #
 ########################################################################
 
+def updateKeyMap(keyboardEvent):
+    """Unsupported convenience method to call sad hacks which should go away."""
+
+    global _restoreOrcaKeys
+    if keyboardEvent.isPressedKey():
+        return
+
+    if keyboardEvent.event_string in settings.orcaModifierKeys \
+       and orca_state.bypassNextCommand:
+        _restoreXmodmap()
+        _restoreOrcaKeys = True
+        return
+
+    if _restoreOrcaKeys and not orca_state.bypassNextCommand:
+        _createOrcaXmodmap()
+        _restoreOrcaKeys = False
+
 def _setXmodmap(xkbmap):
     """Set the keyboard map using xkbcomp."""
     p = subprocess.Popen(['xkbcomp', '-w0', '-', os.environ['DISPLAY']],
@@ -330,8 +255,7 @@ def _setCapsLockAsOrcaModifier(enable):
     capsModLine =    '        action= SetMods(modifiers=Lock,clearLocks);'
     lines = _originalXmodmap.decode('UTF-8').split('\n')
     foundCapsInterpretSection = False
-    for i in range(len(lines)):
-        line = lines[i]
+    for i, line in enumerate(lines):
         if not foundCapsInterpretSection:
             if interpretCapsLineProg.match(line):
                 foundCapsInterpretSection = True
@@ -396,12 +320,14 @@ def loadUserSettings(script=None, inputEvent=None, skipReloadMessage=False):
     Returns True to indicate the input event has been consumed.
     """
 
-    debug.println(debug.LEVEL_FINEST, 'INFO: Loading User Settings')
+    debug.println(debug.LEVEL_INFO, 'ORCA: Loading User Settings', True)
 
     global _userSettings
 
     # Shutdown the output drivers and give them a chance to die.
 
+    player = sound.getPlayer()
+    player.shutdown()
     speech.shutdown()
     braille.shutdown()
 
@@ -412,9 +338,10 @@ def loadUserSettings(script=None, inputEvent=None, skipReloadMessage=False):
         _profile = _settingsManager.getSetting('activeProfile')[1]
         try:
             _userSettings = _settingsManager.getGeneralSettings(_profile)
+            _settingsManager.setProfile(_profile)
             reloaded = True
         except ImportError:
-            debug.printException(debug.LEVEL_FINEST)
+            debug.printException(debug.LEVEL_INFO)
         except:
             debug.printException(debug.LEVEL_SEVERE)
     else:
@@ -422,40 +349,36 @@ def loadUserSettings(script=None, inputEvent=None, skipReloadMessage=False):
         try:
             _userSettings = _settingsManager.getGeneralSettings(_profile)
         except ImportError:
-            debug.printException(debug.LEVEL_FINEST)
+            debug.printException(debug.LEVEL_INFO)
         except:
             debug.printException(debug.LEVEL_SEVERE)
 
+    if not script:
+        script = _scriptManager.getDefaultScript()
+
+    _settingsManager.loadAppSettings(script)
+
     if _settingsManager.getSetting('enableSpeech'):
         try:
             speech.init()
             if reloaded and not skipReloadMessage:
-                speech.speak(messages.SETTINGS_RELOADED,
-                             settings.voices.get(settings.SYSTEM_VOICE))
-            debug.println(debug.LEVEL_CONFIGURATION,
-                          "Speech module has been initialized.")
+                script.speakMessage(messages.SETTINGS_RELOADED)
         except:
             debug.printException(debug.LEVEL_SEVERE)
-            debug.println(debug.LEVEL_SEVERE,
-                          "Could not initialize connection to speech.")
     else:
-        debug.println(debug.LEVEL_CONFIGURATION,
-                      "Speech module has NOT been initialized.")
+        msg = 'ORCA: Speech is not enabled in settings'
+        debug.println(debug.LEVEL_INFO, msg, True)
 
     if _settingsManager.getSetting('enableBraille'):
         try:
             braille.init(_processBrailleEvent, settings.tty)
         except:
             debug.printException(debug.LEVEL_WARNING)
-            debug.println(debug.LEVEL_WARNING,
-                          "Could not initialize connection to braille.")
+            msg = 'ORCA: Could not initialize connection to braille.'
+            debug.println(debug.LEVEL_WARNING, msg, True)
 
-    # I'm not sure where else this should go. But it doesn't really look
-    # right here.
-    try:
-        mouse_review.mouse_reviewer.toggle(on=settings.enableMouseReview)
-    except NameError:
-        pass
+    if _settingsManager.getSetting('enableSound'):
+        player.init()
 
     global _orcaModifiers
     custom = [k for k in settings.orcaModifierKeys if k not in _orcaModifiers]
@@ -471,10 +394,30 @@ def loadUserSettings(script=None, inputEvent=None, skipReloadMessage=False):
     _scriptManager.activate()
     _eventManager.activate()
 
-    debug.println(debug.LEVEL_FINEST, 'INFO: User Settings Loaded')
+    debug.println(debug.LEVEL_INFO, 'ORCA: User Settings Loaded', True)
 
     return True
 
+def _showPreferencesUI(script, prefs):
+    if orca_state.orcaOS:
+        orca_state.orcaOS.showGUI()
+        return
+
+    try:
+        module = importlib.import_module('.orca_gui_prefs', 'orca')
+    except:
+        debug.printException(debug.LEVEL_SEVERE)
+        return
+
+    uiFile = os.path.join(orca_platform.datadir,
+                          orca_platform.package,
+                          "ui",
+                          "orca-setup.ui")
+
+    orca_state.orcaOS = module.OrcaSetupGUI(uiFile, "orcaSetupWindow", prefs)
+    orca_state.orcaOS.init(script)
+    orca_state.orcaOS.showGUI()
+
 def showAppPreferencesGUI(script=None, inputEvent=None):
     """Displays the user interace to configure the settings for a
     specific applications within Orca and set up those app-specific
@@ -483,11 +426,12 @@ def showAppPreferencesGUI(script=None, inputEvent=None):
     Returns True to indicate the input event has been consumed.
     """
 
-    try:
-        module = importlib.import_module('.app_gui_prefs', 'orca')
-        module.showPreferencesUI()
-    except:
-        debug.printException(debug.LEVEL_SEVERE)
+    prefs = {}
+    for key in settings.userCustomizableSettings:
+        prefs[key] = _settingsManager.getSetting(key)
+
+    script = script or orca_state.activeScript
+    _showPreferencesUI(script, prefs)
 
     return True
 
@@ -498,11 +442,9 @@ def showPreferencesGUI(script=None, inputEvent=None):
     Returns True to indicate the input event has been consumed.
     """
 
-    try:
-        module = importlib.import_module('.orca_gui_prefs', 'orca')
-        module.showPreferencesUI()
-    except:
-        debug.printException(debug.LEVEL_SEVERE)
+    prefs = _settingsManager.getGeneralSettings(_settingsManager.profile)
+    script = _scriptManager.getDefaultScript()
+    _showPreferencesUI(script, prefs)
 
     return True
 
@@ -511,6 +453,7 @@ def helpForOrca(script=None, inputEvent=None, page=""):
 
     Returns True to indicate the input event has been consumed.
     """
+    orca_state.learnModeEnabled = False
     uri = "help:orca"
     if page:
         uri += "?%s" % page
@@ -555,7 +498,7 @@ def init(registry):
     module has already been initialized.
     """
 
-    debug.println(debug.LEVEL_FINEST, 'INFO: Initializing Orca module')
+    debug.println(debug.LEVEL_INFO, 'ORCA: Initializing', True)
 
     global _initialized
 
@@ -569,7 +512,6 @@ def init(registry):
         signal.alarm(settings.timeoutTime)
 
     loadUserSettings()
-    _eventManager.registerKeystrokeListener(_processKeyboardEvent)
 
     if settings.timeoutCallback and (settings.timeoutTime > 0):
         signal.alarm(0)
@@ -581,15 +523,15 @@ def init(registry):
     if a11yAppSettings:
         a11yAppSettings.connect('changed', onEnabledChanged)
 
-    debug.println(debug.LEVEL_FINEST, 'INFO: Orca module initialized')
+    debug.println(debug.LEVEL_INFO, 'ORCA: Initialized', True)
 
     return True
 
-def start(registry):
+def start(registry, cacheValues):
     """Starts Orca.
     """
 
-    debug.println(debug.LEVEL_FINEST, 'INFO: Starting Orca')
+    debug.println(debug.LEVEL_INFO, 'ORCA: Starting', True)
 
     if not _initialized:
         init(registry)
@@ -603,11 +545,20 @@ def start(registry):
     if settings.timeoutCallback and (settings.timeoutTime > 0):
         signal.alarm(0)
 
-    if settings.cacheValues:
+    if cacheValues:
         pyatspi.setCacheLevel(pyatspi.CACHE_PROPERTIES)
 
-    debug.println(debug.LEVEL_FINEST, 'INFO: Orca starting registry')
-    registry.start(gil=settings.useGILIdleHandler)
+    try:
+        Gdk.notify_startup_complete()
+    except:
+        msg = 'ERROR: Exception calling Gdk.notification_startup_complete()'
+        debug.println(debug.LEVEL_INFO, msg, True)
+    else:
+        msg = 'ORCA: Startup complete notification made'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
+    debug.println(debug.LEVEL_INFO, 'ORCA: Starting registry', True)
+    registry.start(gil=False)
 
 def die(exitCode=1):
     pid = os.getpid()
@@ -622,8 +573,8 @@ def die(exitCode=1):
         os.kill(pid, signal.SIGTERM)
 
 def timeout(signum=None, frame=None):
-    debug.println(debug.LEVEL_SEVERE,
-                  "TIMEOUT: something has hung.  Aborting.")
+    msg = 'TIMEOUT: something has hung. Aborting.'
+    debug.println(debug.LEVEL_SEVERE, msg, True)
     debug.printStack(debug.LEVEL_ALL)
     debug.examineProcesses()
     die(EXIT_CODE_HANG)
@@ -635,7 +586,7 @@ def shutdown(script=None, inputEvent=None):
     was never initialized.
     """
 
-    debug.println(debug.LEVEL_FINEST, 'INFO: Shutting down Orca')
+    debug.println(debug.LEVEL_INFO, 'ORCA: Shutting down', True)
 
     global _initialized
 
@@ -650,8 +601,8 @@ def shutdown(script=None, inputEvent=None):
 
     orca_state.activeScript.presentMessage(messages.STOP_ORCA)
 
-    _eventManager.deactivate()
     _scriptManager.deactivate()
+    _eventManager.deactivate()
 
     # Shutdown all the other support.
     #
@@ -659,6 +610,9 @@ def shutdown(script=None, inputEvent=None):
         speech.shutdown()
     if settings.enableBraille:
         braille.shutdown()
+    if settings.enableSound:
+        player = sound.getPlayer()
+        player.shutdown()
 
     if settings.timeoutCallback and (settings.timeoutTime > 0):
         signal.alarm(0)
@@ -666,10 +620,10 @@ def shutdown(script=None, inputEvent=None):
     _initialized = False
     _restoreXmodmap(_orcaModifiers)
 
-    debug.println(debug.LEVEL_FINEST, 'INFO: Orca stopping registry')
+    debug.println(debug.LEVEL_INFO, 'ORCA: Stopping registry', True)
     pyatspi.Registry.stop()
 
-    debug.println(debug.LEVEL_FINEST, 'INFO: Orca shutdown complete')
+    debug.println(debug.LEVEL_INFO, 'ORCA: Shutdown complete', True)
 
     return True
 
@@ -677,12 +631,8 @@ exitCount = 0
 def shutdownOnSignal(signum, frame):
     global exitCount
 
-    debug.println(debug.LEVEL_ALL,
-                  "Shutting down and exiting due to signal = %d" \
-                  % signum)
-
-    debug.println(debug.LEVEL_ALL, "Current stack is:")
-    debug.printStack(debug.LEVEL_ALL)
+    msg = 'ORCA: Shutting down and exiting due to signal=%d' % signum
+    debug.println(debug.LEVEL_INFO, msg, True)
 
     # Well...we'll try to exit nicely, but if we keep getting called,
     # something bad is happening, so just quit.
@@ -717,19 +667,18 @@ def shutdownOnSignal(signum, frame):
     if not cleanExit:
         die(EXIT_CODE_HANG)
 
-def abortOnSignal(signum, frame):
-    debug.println(debug.LEVEL_ALL,
-                  "Aborting due to signal = %d" \
-                  % signum)
-    die(signum)
-
-def main():
+def main(cacheValues=True):
     """The main entry point for Orca.  The exit codes for Orca will
     loosely be based on signals, where the exit code will be the
     signal used to terminate Orca (if a signal was used).  Otherwise,
     an exit code of 0 means normal completion and an exit code of 50
     means Orca exited because of a hang."""
 
+    if debug.debugFile and os.path.exists(debug.debugFile.name):
+        faulthandler.enable(file=debug.debugFile, all_threads=False)
+    else:
+        faulthandler.enable(all_threads=False)
+
     # Method to call when we think something might be hung.
     #
     settings.timeoutCallback = timeout
@@ -740,7 +689,6 @@ def main():
     signal.signal(signal.SIGINT, shutdownOnSignal)
     signal.signal(signal.SIGTERM, shutdownOnSignal)
     signal.signal(signal.SIGQUIT, shutdownOnSignal)
-    signal.signal(signal.SIGSEGV, abortOnSignal)
 
     if not _settingsManager.isAccessibilityEnabled():
         _settingsManager.setAccessibility(True)
@@ -749,11 +697,8 @@ def main():
 
     try:
         message = messages.START_ORCA
-        if not _settingsManager.getSetting('onlySpeakDisplayedText'):
-            speech.speak(message, settings.voices.get(settings.SYSTEM_VOICE))
-        if _settingsManager.getSetting('enableBraille') \
-           or _settingsManager.getSetting('enableBrailleMonitor'):
-            braille.displayMessage(message)
+        script = _scriptManager.getDefaultScript()
+        script.presentMessage(message)
     except:
         debug.printException(debug.LEVEL_SEVERE)
 
@@ -764,7 +709,7 @@ def main():
             setLocusOfFocus(None, window)
 
     try:
-        start(pyatspi.Registry) # waits until we stop the registry
+        start(pyatspi.Registry, cacheValues) # waits until we stop the registry
     except:
         die(EXIT_CODE_HANG)
     return 0
