@@ -31,6 +31,7 @@ __license__   = "LGPL"
 import pyatspi
 
 import orca.debug as debug
+import orca.keybindings as keybindings
 import orca.orca_state as orca_state
 import orca.script_utilities as script_utilities
 
@@ -84,7 +85,7 @@ class Utilities(script_utilities.Utilities):
                 return text
 
         try:
-            text = script_utilities.Utilities.displayedText(self, obj)
+            text = super().displayedText(obj)
         except:
             return ""
 
@@ -92,71 +93,38 @@ class Utilities(script_utilities.Utilities):
         # back on the name, which is bogus. Once that has been fixed, this
         # hack can go.
         if role == pyatspi.ROLE_TABLE_CELL and text == obj.name \
-           and (self.isSpreadSheetCell(obj) or self.isDocumentCell(obj)):
+           and (self.isSpreadSheetCell(obj) or self.isTextDocumentCell(obj)):
             return ""
 
+        # More bogusness from (at least) Calc combined with the aforementioned
+        # fallback-to-name behavior....
+        if self.isDocument(obj) and text == obj.name and obj.name.startswith("file:///"):
+            return ""
+
+        # TODO - JD: Once the VCL script is completed and subclasses the
+        # appropriate toolkit scripts, this should not be needed.
+        if obj.parent and obj.parent.getRole() == pyatspi.ROLE_LIST_BOX:
+            labels = self.unrelatedLabels(obj, onlyShowing=False)
+            text = " ".join(map(self.displayedText, labels))
+
         return text
 
-    def isReadOnlyTextArea(self, obj):
-        """Returns True if obj is a text entry area that is read only."""
-
-        if not obj.getRole() == pyatspi.ROLE_TEXT:
-            return False
-
-        state = obj.getState()
-        readOnly = state.contains(pyatspi.STATE_FOCUSABLE) \
-                   and not state.contains(pyatspi.STATE_EDITABLE)
-        details = debug.getAccessibleDetails(debug.LEVEL_ALL, obj)
-        debug.println(debug.LEVEL_ALL,
-                      "soffice - isReadOnlyTextArea=%s for %s" % \
-                      (readOnly, details))
-
-        return readOnly
-
     def isCellBeingEdited(self, obj):
         if not obj:
             return False
 
         parent = obj.parent
-        if parent and parent.getRoleName() == 'text frame':
-            if self.spreadSheetCellName(parent):
-                return True
-
-        return False
-
-    def isSpreadSheetCell(self, obj, startFromTable=False):
-        """Return an indication of whether the given obj is a spread sheet
-        table cell.
-
-        Arguments:
-        - obj: the object to check.
-        - startFromTable: if True, then the component hierarchy check should
-          start from a table (as opposed to a table cell).
-
-        Returns True if this is a table cell, False otherwise.
-        """
-
-        cell = obj
-        if not startFromTable:
-            obj = obj.parent
-
         try:
-            table = obj.queryTable()
+            role = parent.getRole()
         except:
-            return self.isCellBeingEdited(cell)
-        else:
-            return table.nRows in [65536, 1048576]
-
-    def isDocumentCell(self, cell):
-        isCell = lambda x: x and x.getRole() == pyatspi.ROLE_TABLE_CELL
-        if not isCell(cell):
-            cell = pyatspi.findAncestor(cell, isCell)
-
-        if not cell or self.isSpreadSheetCell(cell):
+            msg = "SOFFICE: Exception getting role of %s" % parent
+            debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
-        isDocument = lambda x: x and x.getRole() == pyatspi.ROLE_DOCUMENT_FRAME
-        return pyatspi.findAncestor(cell, isDocument) != None
+        if role in [pyatspi.ROLE_EXTENDED, pyatspi.ROLE_PANEL]:
+            return self.spreadSheetCellName(parent)
+
+        return False
 
     def spreadSheetCellName(self, cell):
         nameList = cell.name.split()
@@ -193,66 +161,19 @@ class Utilities(script_utilities.Utilities):
 
         return row, column, table
 
-    def getShowingCellsInRow(self, obj):
-        row, column, parentTable = self.getRowColumnAndTable(obj)
-        try:
-            table = parentTable.queryTable()
-        except:
-            return []
+    def rowHeadersForCell(self, obj):
+        rowHeader, colHeader = self.getDynamicHeadersForCell(obj)
+        if rowHeader:
+            return [rowHeader]
 
-        startIndex, endIndex = self.getTableRowRange(obj)
-        cells = []
-        for i in range(startIndex, endIndex):
-            cell = table.getAccessibleAt(row, i)
-            try:
-                showing = cell.getState().contains(pyatspi.STATE_SHOWING)
-            except:
-                continue
-            if showing:
-                cells.append(cell)
+        return super().rowHeadersForCell(obj)
 
-        return cells
+    def columnHeadersForCell(self, obj):
+        rowHeader, colHeader = self.getDynamicHeadersForCell(obj)
+        if colHeader:
+            return [colHeader]
 
-    def getTableRowRange(self, obj):
-        """If this is spread sheet cell, return the start and end indices
-        of the spread sheet cells for the table that obj is in. Otherwise
-        return the complete range (0, parentTable.nColumns).
-
-        Arguments:
-        - obj: a table cell.
-
-        Returns the start and end table cell indices.
-        """
-
-        parent = obj.parent
-        try:
-            parentTable = parent.queryTable()
-        except NotImplementedError:
-            parentTable = None
-
-        startIndex = 0
-        endIndex = parentTable.nColumns
-
-        if self.isSpreadSheetCell(obj):
-            extents = parent.queryComponent().getExtents(pyatspi.DESKTOP_COORDS)
-            y = extents.y
-            leftX = extents.x + 1
-            leftCell = \
-                parent.queryComponent().getAccessibleAtPoint(leftX, y, 0)
-            if leftCell:
-                table = leftCell.parent.queryTable()
-                index = self.cellIndex(leftCell)
-                startIndex = table.getColumnAtIndex(index)
-
-            rightX = extents.x + extents.width - 1
-            rightCell = \
-                parent.queryComponent().getAccessibleAtPoint(rightX, y, 0)
-            if rightCell:
-                table = rightCell.parent.queryTable()
-                index = self.cellIndex(rightCell)
-                endIndex = table.getColumnAtIndex(index) + 1
-
-        return [startIndex, endIndex]
+        return super().columnHeadersForCell(obj)
 
     def getDynamicHeadersForCell(self, obj, onlyIfNew=False):
         if not (self._script.dynamicRowHeaders or self._script.dynamicColumnHeaders):
@@ -267,8 +188,8 @@ class Utilities(script_utilities.Utilities):
         if headersRow == objRow or headersCol == objCol:
             return None, None
 
-        getRowHeader = headersCol != None
-        getColHeader = headersRow != None
+        getRowHeader = headersCol is not None
+        getColHeader = headersRow is not None
         if onlyIfNew:
             getRowHeader = \
                 getRowHeader and objRow != self._script.pointOfReference.get("lastRow")
@@ -285,35 +206,46 @@ class Utilities(script_utilities.Utilities):
 
         return rowHeader, colHeader
 
-    def isSameObject(self, obj1, obj2):
-        same = script_utilities.Utilities.isSameObject(self, obj1, obj2)
-        if not same or obj1 == obj2:
-            return same
+    def isSameObject(self, obj1, obj2, comparePaths=False, ignoreNames=False):
+        if obj1 == obj2:
+            return True
 
-        # The document frame currently contains just the active page,
-        # resulting in false positives. So for paragraphs, rely upon
-        # the equality check.
-        if obj1.getRole() == obj2.getRole() == pyatspi.ROLE_PARAGRAPH:
+        try:
+            role1 = obj1.getRole()
+            role2 = obj2.getRole()
+        except:
             return False
 
-        # Handle the case of false positives in dialog boxes resulting
-        # from getIndexInParent() returning a bogus value. bgo#618790.
-        #
-        if not obj1.name \
-           and obj1.getRole() == pyatspi.ROLE_TABLE_CELL \
-           and obj1.getIndexInParent() == obj2.getIndexInParent() == -1:
-            top = self.topLevelObject(obj1)
-            if top and top.getRole() == pyatspi.ROLE_DIALOG:
-                same = False
+        if role1 != role2 or role1 == pyatspi.ROLE_PARAGRAPH:
+            return False
 
-        return same
+        try:
+            name1 = obj1.name
+            name2 = obj2.name
+        except:
+            return False
+
+        if name1 == name2:
+            if role1 == pyatspi.ROLE_FRAME:
+                return True
+            if role1 == pyatspi.ROLE_TABLE_CELL and not name1:
+                if self.isZombie(obj1) and self.isZombie(obj2):
+                    return False
+
+        return super().isSameObject(obj1, obj2, comparePaths, ignoreNames)
 
     def isLayoutOnly(self, obj):
         """Returns True if the given object is a container which has
         no presentable information (label, name, displayed text, etc.)."""
 
-        role = obj.getRole()
-        if role == pyatspi.ROLE_PANEL and obj.childCount == 1:
+        try:
+            role = obj.getRole()
+            childCount = obj.childCount
+        except:
+            role = None
+            childCount = 0
+
+        if role == pyatspi.ROLE_PANEL and childCount == 1:
             if obj.name and obj.name == obj[0].name:
                 return True
 
@@ -321,7 +253,38 @@ class Utilities(script_utilities.Utilities):
            and obj.parent.getRole() == pyatspi.ROLE_COMBO_BOX:
             return True
 
-        return script_utilities.Utilities.isLayoutOnly(self, obj)
+        return super().isLayoutOnly(obj)
+
+    def isAnInputLine(self, obj):
+        if not obj:
+            return False
+        if obj == self.locateInputLine(obj):
+            return True
+
+        parent = obj.parent
+        try:
+            role = parent.getRole()
+        except:
+            msg = "SOFFICE: Exception getting role of %s" % parent
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return False
+
+        if role in [pyatspi.ROLE_EXTENDED, pyatspi.ROLE_PANEL]:
+            if self.spreadSheetCellName(parent):
+                return False
+
+        parent = parent.parent
+        try:
+            role = parent.getRole()
+        except:
+            msg = "SOFFICE: Exception getting role of %s" % parent
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return False
+
+        if role == pyatspi.ROLE_TEXT:
+            return True
+
+        return False
 
     def locateInputLine(self, obj):
         """Return the spread sheet input line. This only needs to be found
@@ -338,7 +301,15 @@ class Utilities(script_utilities.Utilities):
         """
 
         if self._script.inputLineForCell:
-            return self._script.inputLineForCell
+            try:
+                topLevel = self.topLevelObject(self._script.inputLineForCell)
+            except:
+                msg = "ERROR: Exception getting topLevelObject for inputline"
+                debug.println(debug.LEVEL_INFO, msg, True)
+                self._script.inputLineForCell = None
+            else:
+                if self.isSameObject(orca_state.activeWindow, topLevel):
+                    return self._script.inputLineForCell
 
         isScrollPane = lambda x: x and x.getRole() == pyatspi.ROLE_SCROLL_PANE
         scrollPane = pyatspi.findAncestor(obj, isScrollPane)
@@ -352,8 +323,9 @@ class Utilities(script_utilities.Utilities):
                 break
 
         if not toolbar:
-            debug.println(debug.LEVEL_INFO, "Calc inputline toolbar not found.")
-            return
+            msg = "ERROR: Calc inputline toolbar not found."
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return None
 
         isParagraph = lambda x: x and x.getRole() == pyatspi.ROLE_PARAGRAPH
         allParagraphs = pyatspi.findAllDescendants(toolbar, isParagraph)
@@ -387,35 +359,45 @@ class Utilities(script_utilities.Utilities):
 
         return results
 
-    def isFunctionalDialog(self, obj):
-        """Returns true if the window is functioning as a dialog."""
+    def validatedTopLevelObject(self, obj):
+        # TODO - JD: We cannot just override topLevelObject() because that will
+        # break flat review access to document content in LO using Gtk+ 3. That
+        # bug seems to be fixed in LO v5.3.0. When that version is released, this
+        # and hopefully other hacks can be removed.
+        window = super().topLevelObject(obj)
+        if not window or window.getIndexInParent() >= 0:
+            return window
 
-        # The OOo Navigator window looks like a dialog, acts like a
-        # dialog, and loses focus requiring the user to know that it's
-        # there and needs Alt+F6ing into.  But officially it's a normal
-        # window.
+        msg = "SOFFICE: %s's window %s has -1 indexInParent" % (obj, window)
+        debug.println(debug.LEVEL_INFO, msg, True)
 
-        # There doesn't seem to be (an efficient) top-down equivalent
-        # of utilities.hasMatchingHierarchy(). But OOo documents have
-        # root panes; this thing does not.
-        #
-        rolesList = [pyatspi.ROLE_FRAME,
-                     pyatspi.ROLE_PANEL,
-                     pyatspi.ROLE_PANEL,
-                     pyatspi.ROLE_TOOL_BAR,
-                     pyatspi.ROLE_PUSH_BUTTON]
+        for child in self._script.app:
+            if self.isSameObject(child, window):
+                window = child
+                break
 
-        if obj.getRole() != rolesList[0]:
-            # We might be looking at the child.
-            #
-            rolesList.pop(0)
+        try:
+            index = window.getIndexInParent()
+        except:
+            index = -1
 
-        while obj and obj.childCount and len(rolesList):
-            if obj.getRole() != rolesList.pop(0):
-                return False
-            obj = obj[0]
+        msg = "SOFFICE: Returning %s (index: %i)" % (window, index)
+        debug.println(debug.LEVEL_INFO, msg, True)
+        return window
 
-        return True
+    def commonAncestor(self, a, b):
+        ancestor = super().commonAncestor(a, b)
+        if ancestor or not (a and b):
+            return ancestor
+
+        windowA = self.validatedTopLevelObject(a)
+        windowB = self.validatedTopLevelObject(b)
+        if not self.isSameObject(windowA, windowB):
+            return None
+
+        msg = "SOFFICE: Adjusted ancestor %s and %s to %s" % (a, b, windowA)
+        debug.println(debug.LEVEL_INFO, msg, True)
+        return windowA
 
     def validParent(self, obj):
         """Returns the first valid parent/ancestor of obj. We need to do
@@ -445,50 +427,6 @@ class Utilities(script_utilities.Utilities):
 
         return parent
 
-    def findPreviousObject(self, obj):
-        """Finds the object before this one."""
-
-        if not obj:
-            return None
-
-        for relation in obj.getRelationSet():
-            if relation.getRelationType() == pyatspi.RELATION_FLOWS_FROM:
-                return relation.getTarget(0)
-
-        index = obj.getIndexInParent() - 1
-        if not (0 <= index < obj.parent.childCount - 1):
-            obj = obj.parent
-            index = obj.getIndexInParent() - 1
-
-        try:
-            prevObj = obj.parent[index]
-        except:
-            prevObj = obj
-
-        return prevObj
-
-    def findNextObject(self, obj):
-        """Finds the object after this one."""
-
-        if not obj:
-            return None
-
-        for relation in obj.getRelationSet():
-            if relation.getRelationType() == pyatspi.RELATION_FLOWS_TO:
-                return relation.getTarget(0)
-
-        index = obj.getIndexInParent() + 1
-        if not (0 < index < obj.parent.childCount):
-            obj = obj.parent
-            index = obj.getIndexInParent() + 1
-
-        try:
-            nextObj = obj.parent[index]
-        except:
-            nextObj = None
-
-        return nextObj
-
     @staticmethod
     def _flowsFromOrToSelection(obj):
         try:
@@ -509,6 +447,18 @@ class Utilities(script_utilities.Utilities):
 
         return False
 
+    def objectContentsAreInClipboard(self, obj=None):
+        obj = obj or orca_state.locusOfFocus
+        if not obj:
+            return False
+
+        if self.isSpreadSheetCell(obj):
+            contents = self.getClipboardContents()
+            string = self.displayedText(obj) or "\n"
+            return string in contents
+
+        return super().objectContentsAreInClipboard(obj)
+
     #########################################################################
     #                                                                       #
     # Impress-Specific Utilities                                            #
@@ -519,20 +469,12 @@ class Utilities(script_utilities.Utilities):
         """Attempts to locate the Impress drawing view, which is the
         area in which slide editing occurs."""
 
-        # TODO - JD: We should probably add this to the generatorCache.
-        #
-        docFrames = self.descendantsWithRole(
-            self.topLevelObject(obj), pyatspi.ROLE_DOCUMENT_FRAME)
-        docFrame = [o for o in docFrames if ":" in o.name and "/" in o.name]
-        if docFrame:
-            return docFrame[0]
-
-        return None
+        return pyatspi.findDescendant(self.topLevelObject(obj), self.isDrawingView)
 
     def isDrawingView(self, obj):
         """Returns True if obj is the Impress Drawing View."""
 
-        if obj and obj.getRole() == pyatspi.ROLE_DOCUMENT_FRAME:
+        if self.isDocument(obj):
             return (":" in obj.name and "/" in obj.name)
 
         return False
@@ -544,8 +486,19 @@ class Utilities(script_utilities.Utilities):
         # that the Frame name will end with "Impress", unlocalized.
         #
         if obj:
-            topLevel = self.topLevelObject(obj)
-            if topLevel and topLevel.name.endswith("Impress"):
+            try:
+                topLevel = self.topLevelObject(obj)
+            except:
+                msg = "ERROR: Exception getting top-level object for %s" % obj
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return False
+            if not topLevel:
+                return False
+            if self.isDead(topLevel):
+                msg = "SOFFICE: Top level object %s is dead." % topLevel
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return False
+            if topLevel.name.endswith("Impress"):
                 return True
 
         return False
@@ -563,12 +516,13 @@ class Utilities(script_utilities.Utilities):
         if not parent:
             return None, None
 
-        panes = self.descendantsWithRole(parent, pyatspi.ROLE_SPLIT_PANE)
+        hasRole = lambda x: x and x.getRole() == pyatspi.ROLE_SPLIT_PANE
+        panes = pyatspi.findAllDescendants(parent, hasRole)
         if not panes:
             return None, None
 
         slidePane = taskPane = None
-        if self.descendantsWithRole(panes[0], pyatspi.ROLE_DOCUMENT_FRAME):
+        if pyatspi.findAllDescendants(panes[0], self.isDocument):
             slidePane = panes[0]
             if len(panes) == 2:
                 taskPane = panes[1]
@@ -586,10 +540,10 @@ class Utilities(script_utilities.Utilities):
         Returns a (title, position, count) tuple.
         """
 
-        if obj.getRole() == pyatspi.ROLE_DOCUMENT_FRAME:
+        if self.isDocument(obj):
             dv = obj
         else:
-            dv = self.ancestorWithRole(obj, [pyatspi.ROLE_DOCUMENT_FRAME], [])
+            dv = self.getContainingDocument(obj)
 
         if not dv or not self.isDrawingView(dv):
             return "", 0, 0
@@ -615,64 +569,6 @@ class Utilities(script_utilities.Utilities):
 
         return title, int(position), int(count)
 
-    #########################################################################
-    #                                                                       #
-    # Utilities for working with the accessible text interface              #
-    #                                                                       #
-    #########################################################################
-
-    def isWordMisspelled(self, obj, offset):
-        """Identifies if the current word is flagged as misspelled by the
-        application.
-
-        Arguments:
-        - obj: An accessible which implements the accessible text interface.
-        - offset: Offset in the accessible's text for which to retrieve the
-          attributes.
-
-        Returns True if the word is flagged as misspelled.
-        """
-
-        attributes, start, end  = self.textAttributes(obj, offset, True)
-        error = attributes.get("text-spelling")
-
-        return error == "misspelled"
-
-    def textAttributes(self, acc, offset, get_defaults=False):
-        """Get the text attributes run for a given offset in a given accessible
-
-        Arguments:
-        - acc: An accessible.
-        - offset: Offset in the accessible's text for which to retrieve the
-        attributes.
-        - get_defaults: Get the default attributes as well as the unique ones.
-        Default is True
-
-        Returns a dictionary of attributes, a start offset where the attributes
-        begin, and an end offset. Returns ({}, 0, 0) if the accessible does not
-        supprt the text attribute.
-        """
-        rv, start, end = script_utilities.Utilities.\
-            textAttributes(self, acc, offset, get_defaults)
-
-        # If there are no text attributes associated with the text at a
-        # given offset, we might get some seriously bogus offsets, in
-        # particular, an extremely large start offset and an extremely
-        # large, but negative end offset. As a result, any text attributes
-        # which are present on the line after the specified offset will
-        # not be indicated by braille.py's getAttributeMask. Therefore,
-        # we'll set the start offset to the character being examined,
-        # and the end offset to the next character.
-        #
-        start = min(start, offset)
-        if end < 0:
-            debug.println(debug.LEVEL_WARNING,
-                "soffice.script.py:getTextAttributes: detected a bogus " +
-                "end offset. Start offset: %s, end offset: %s" % (start, end))
-            end = offset + 1
-
-        return rv, start, end
-
     #########################################################################
     #                                                                       #
     # Miscellaneous Utilities                                               #
@@ -711,15 +607,93 @@ class Utilities(script_utilities.Utilities):
 
         return False
 
+    def containingComboBox(self, obj):
+        isComboBox = lambda x: x and x.getRole() == pyatspi.ROLE_COMBO_BOX
+        if isComboBox(obj):
+            comboBox = obj
+        else:
+            comboBox = pyatspi.findAncestor(obj, isComboBox)
+
+        if not comboBox:
+            return None
+
+        if not self.isZombie(comboBox):
+            return comboBox
+
+        try:
+            parent = comboBox.parent
+        except:
+            pass
+        else:
+            replicant = self.findReplicant(parent, comboBox)
+            if replicant and not self.isZombie(replicant):
+                comboBox = replicant
+
+        return comboBox
+
+    def isComboBoxSelectionChange(self, event):
+        comboBox = self.containingComboBox(event.source)
+        if not comboBox:
+            return False
+
+        lastKey, mods = self.lastKeyAndModifiers()
+        if not lastKey in ["Down", "Up"]:
+            return False
+
+        return True
+
+    def isComboBoxNoise(self, event):
+        role = event.source.getRole()
+        if role == pyatspi.ROLE_TEXT and event.type.startswith("object:text-"):
+            return self.isComboBoxSelectionChange(event)
+
+        return False
+
+    def isPresentableTextChangedEventForLocusOfFocus(self, event):
+        if self.isComboBoxNoise(event):
+            msg = "SOFFICE: Event is believed to be combo box noise"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return False
+
+        return super().isPresentableTextChangedEventForLocusOfFocus(event)
+
+    def isReadOnlyTextArea(self, obj):
+        if not super().isReadOnlyTextArea(obj):
+            return False
+
+        return not self.inDocumentContent(obj)
+
+    def isSelectedTextDeletionEvent(self, event):
+        if event.type.startswith("object:state-changed:selected") and not event.detail1:
+            return self.isDead(orca_state.locusOfFocus) and self.lastInputEventWasDelete()
+
+        return super().isSelectedTextDeletionEvent(event)
+
+    def lastInputEventWasRedo(self):
+        if super().lastInputEventWasRedo():
+            return True
+
+        keyString, mods = self.lastKeyAndModifiers()
+        if mods & keybindings.COMMAND_MODIFIER_MASK and keyString.lower() == 'y':
+            return not (mods & keybindings.SHIFT_MODIFIER_MASK)
+
+        return False
+
     def selectedChildren(self, obj):
         if not obj:
             return []
 
+        role = obj.getRole()
+        isSelection = lambda x: x and 'Selection' in pyatspi.listInterfaces(x)
+        if not isSelection(obj) and role == pyatspi.ROLE_COMBO_BOX:
+            child = pyatspi.findDescendant(obj, isSelection)
+            if child:
+                return super().selectedChildren(child)
+
         # Things only seem broken for certain tables, e.g. the Paths table.
         # TODO - JD: File the LibreOffice bugs and reference them here.
-        if obj.getRole() != pyatspi.ROLE_TABLE \
-           or self.isSpreadSheetCell(obj, True):
-            return script_utilities.Utilities.selectedChildren(self, obj)
+        if role != pyatspi.ROLE_TABLE or self.isSpreadSheetCell(obj):
+            return super().selectedChildren(obj)
 
         try:
             selection = obj.querySelection()
@@ -732,3 +706,25 @@ class Utilities(script_utilities.Utilities):
                 children.append(obj[i])
 
         return children
+
+    def getFirstCaretPosition(self, obj):
+        try:
+            text = obj.queryText()
+        except:
+            if obj and obj.childCount:
+                return self.getFirstCaretPosition(obj[0])
+
+        return obj, 0
+
+    def shouldReadFullRow(self, obj):
+        if self._script._lastCommandWasStructNav:
+            return False
+
+        lastKey, mods = self.lastKeyAndModifiers()
+        if lastKey in ["Tab", "ISO_Left_Tab"]:
+            return False
+
+        if not super().shouldReadFullRow(obj):
+            return False
+
+        return self.cellRowChanged(obj)
