@@ -170,7 +170,7 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         """
 
         result = []
-        if not self._script.utilities.isSpreadSheetCell(obj, startFromTable=True):
+        if not self._script.utilities.isSpreadSheetCell(obj):
             result.extend(speech_generator.SpeechGenerator.\
                 _generateAvailability(self, obj, **args))
 
@@ -209,7 +209,7 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return result
 
     def _generateCurrentLineText(self, obj, **args):
-        if self._script.utilities.isDocumentCell(obj.parent):
+        if self._script.utilities.isTextDocumentCell(obj.parent):
             priorObj = args.get('priorObj', None)
             if priorObj and priorObj.parent != obj.parent:
                 return []
@@ -218,12 +218,16 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         # together in the scripts. In addition, the voice crap needs to go
         # here. Then it needs to be removed from the scripts.
         [text, caretOffset, startOffset] = self._script.getTextLineAtCaret(obj)
+        voice = self.voice(string=text)
         text = self._script.utilities.adjustForLinks(obj, text, startOffset)
         text = self._script.utilities.adjustForRepeats(text)
         if not text:
-            text = [messages.BLANK]
+            result = [messages.BLANK]
+        else:
+            result = [text]
+        result.extend(voice)
 
-        return [text]
+        return result
 
     def _generateToggleState(self, obj, **args):
         """Treat toggle buttons in the toolbar specially. This is so we can
@@ -251,7 +255,7 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         row header(s).
         """
 
-        if _settingsManager.getSetting('readTableCellRow'):
+        if self._script.utilities.shouldReadFullRow(obj):
             return []
 
         newOnly = args.get('newOnly', False)
@@ -357,10 +361,12 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         Returns a list of utterances to be spoken for the object.
         """
 
-        result = speech_generator.SpeechGenerator._generateRealTableCell(
-            self, obj, **args)
+        result = super()._generateRealTableCell(obj, **args)
 
         if not self._script.utilities.isSpreadSheetCell(obj):
+            if self._script._lastCommandWasStructNav:
+                return result
+
             if _settingsManager.getSetting('speakCellCoordinates'):
                 result.append(obj.name)
             return result
@@ -370,7 +376,7 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         if speakCoordinates and not isBasicWhereAmI:
             result.append(self._script.utilities.spreadSheetCellName(obj))
 
-        if _settingsManager.getSetting('readTableCellRow'):
+        if self._script.utilities.shouldReadFullRow(obj):
             row, col, table = self._script.utilities.getRowColumnAndTable(obj)
             lastRow = self._script.pointOfReference.get("lastRow")
             if row != lastRow:
@@ -389,24 +395,17 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return result
 
     def _generateTableCellRow(self, obj, **args):
-        """Get the speech for a table cell row if the user wants to hear
-        the full row and if the row has actually changed."""
-
-        speakFullRow = _settingsManager.getSetting('readTableCellRow')
-        if speakFullRow:
-            row, column, table = \
-                self._script.utilities.getRowColumnAndTable(obj)
-            lastRow = self._script.pointOfReference.get("lastRow")
-            if lastRow == -1 and self._script.utilities.isSpreadSheetCell(obj):
-                speakFullRow = False
-            else:
-                speakFullRow = row != lastRow
-
-        if not speakFullRow:
+        if not self._script.utilities.shouldReadFullRow(obj):
             return self._generateRealTableCell(obj, **args)
 
+        if not self._script.utilities.isSpreadSheetCell(obj):
+            return super()._generateTableCellRow(obj, **args)
+
+        cells = self._script.utilities.getShowingCellsInSameRow(obj)
+        if not cells:
+            return []
+
         result = []
-        cells = self._script.utilities.getShowingCellsInRow(obj)
         for cell in cells:
             result.extend(self._generateRealTableCell(cell, **args))
 
@@ -420,6 +419,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         dialog).
         """
 
+        if self._script._lastCommandWasStructNav:
+            return []
+
         topLevel = self._script.utilities.topLevelObject(obj)
         if topLevel and topLevel.getRole() == pyatspi.ROLE_DIALOG:
             return []
@@ -444,14 +446,14 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         if not priorObj:
             return []
 
-        if obj.getRoleName() == 'text frame':
+        if self._script.utilities.isSpreadSheetCell(priorObj):
             return []
 
         isTable = lambda x: x and x.getRole() == pyatspi.ROLE_TABLE
         oldTable = pyatspi.findAncestor(priorObj, isTable)
         if oldTable:
             ancestor = self._script.utilities.commonAncestor(oldTable, obj)
-            if ancestor and ancestor.getRole() == pyatspi.ROLE_DOCUMENT_FRAME:
+            if self._script.utilities.isDocument(ancestor):
                 result = [messages.TABLE_LEAVING]
                 result.extend(self.voice(speech_generator.SYSTEM))
                 result.extend(self._generatePause(obj, **args))
@@ -464,6 +466,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         if self._script.utilities.isSpreadSheetCell(obj):
             return []
 
+        if self._script._lastCommandWasStructNav:
+            return []
+
         return super()._generateUnselectedCell(obj, **args)
 
     def generateSpeech(self, obj, **args):
@@ -480,11 +485,12 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
             # we can use to guess the coordinates.
             #
             args['guessCoordinates'] = obj.getRole() == pyatspi.ROLE_PARAGRAPH
-            result.extend(speech_generator.SpeechGenerator.\
-                                           generateSpeech(self, obj, **args))
+            result.extend(super().generateSpeech(obj, **args))
             del args['guessCoordinates']
             self._restoreRole(oldRole, args)
         else:
-            result.extend(speech_generator.SpeechGenerator.\
-                                           generateSpeech(self, obj, **args))
+            oldRole = self._overrideRole(self._getAlternativeRole(obj, **args), args)
+            result.extend(super().generateSpeech(obj, **args))
+            self._restoreRole(oldRole, args)
+
         return result
