@@ -27,7 +27,6 @@ __license__   = "LGPL"
 
 import pyatspi
 import urllib.parse, urllib.request, urllib.error, urllib.parse
-from gi.repository import Atspi, Atk
 
 from . import chnames
 from . import debug
@@ -210,6 +209,10 @@ class SpeechGenerator(generator.Generator):
         if _settingsManager.getSetting('onlySpeakDisplayedText'):
             return []
 
+        priorObj = args.get('priorObj')
+        if priorObj and priorObj.getRole() == pyatspi.ROLE_TOOL_TIP:
+            return []
+
         acss = self.voice(SYSTEM)
         result = generator.Generator._generateDescription(self, obj, **args)
         if result:
@@ -345,12 +348,12 @@ class SpeechGenerator(generator.Generator):
             level = self._script.utilities.headingLevel(obj)
             if level:
                 result.append(object_properties.ROLE_HEADING_LEVEL_SPEECH % {
-                    'role': self.getLocalizedRoleName(obj, role),
+                    'role': self.getLocalizedRoleName(obj, **args),
                     'level': level})
                 result.extend(acss)
 
         if role not in doNotPresent and not result:
-            result.append(self.getLocalizedRoleName(obj, role))
+            result.append(self.getLocalizedRoleName(obj, **args))
             result.extend(acss)
         return result
 
@@ -362,30 +365,37 @@ class SpeechGenerator(generator.Generator):
         accessible role of the obj.  This is provided mostly as a
         method for scripts to call.
         """
-        return self._generateRoleName(obj, **args)
+        generated = self._generateRoleName(obj, **args)
+        if generated:
+            return generated[0]
 
-    def getLocalizedRoleName(self, obj, role=None):
+        return ""
+
+    def getName(self, obj, **args):
+        generated = self._generateName(obj, **args)
+        if generated:
+            return generated[0]
+
+        return ""
+
+    def getLocalizedRoleName(self, obj, **args):
         """Returns the localized name of the given Accessible object; the name
         is suitable to be spoken.
 
         Arguments:
         - obj: an Accessible object
-        - role: an optional pyatspi role to use instead
         """
 
-        if not isinstance(role, pyatspi.Role):
-            try:
-                return obj.getLocalizedRoleName()
-            except:
-                return ''
+        if self._script.utilities.isEditableComboBox(obj) \
+           or self._script.utilities.isEditableDescendantOfComboBox(obj):
+            return object_properties.ROLE_EDITABLE_COMBO_BOX
 
-        if not role:
-            return ''
+        role = args.get('role', obj.getRole())
+        state = obj.getState()
+        if role == pyatspi.ROLE_LINK and state.contains(pyatspi.STATE_VISITED):
+            return object_properties.ROLE_VISITED_LINK
 
-        nonlocalized = Atspi.role_get_name(role)
-        atkRole = Atk.role_for_name(nonlocalized)
-
-        return Atk.role_get_localized_name(atkRole)
+        return super().getLocalizedRoleName(obj, **args)
 
     def _generateUnrelatedLabels(self, obj, **args):
         """Returns, as an array of strings (and possibly voice
@@ -458,19 +468,15 @@ class SpeechGenerator(generator.Generator):
     def _generateMultiselectableState(self, obj, **args):
         """Returns an array of strings (and possibly voice and audio
         specifications) that represent the multiselectable state of
-        the object.  This is typically for check boxes. If the object
+        the object.  This is typically for list boxes. If the object
         is not multiselectable, an empty array will be returned.
         """
         if _settingsManager.getSetting('onlySpeakDisplayedText'):
             return []
 
-        result = []
         acss = self.voice(STATE)
-        if obj.getState().contains(pyatspi.STATE_MULTISELECTABLE) \
-           and obj.childCount:
-            result.append(self._script.formatting.getString(
-                mode='speech',
-                stringType='multiselect'))
+        result = super()._generateMultiselectableState(obj, **args)
+        if result:
             result.extend(acss)
         return result
 
@@ -510,6 +516,14 @@ class SpeechGenerator(generator.Generator):
     #                                                                   #
     #####################################################################
 
+    def generateLinkInfo(self, obj, **args):
+        result = self._generateLinkInfo(obj, **args)
+        result.extend(self._generatePause(obj, **args))
+        result.append(self._generateSiteDescription(obj, **args))
+        result.extend(self._generatePause(obj, **args))
+        result.append(self._generateFileSize(obj, **args))
+        return result
+
     def _generateLinkInfo(self, obj, **args):
         """Returns an array of strings (and possibly voice and audio
         specifications) that represent the protocol of the URI of
@@ -543,6 +557,14 @@ class SpeechGenerator(generator.Generator):
             else:
                 linkOutput = messages.LINK_WITH_PROTOCOL % link_uri_info[0]
                 text = self._script.utilities.displayedText(obj)
+                try:
+                    isVisited = obj.getState().contains(pyatspi.STATE_VISITED)
+                except:
+                    isVisited = False
+                if not isVisited:
+                    linkOutput = messages.LINK_WITH_PROTOCOL % link_uri_info[0]
+                else:
+                    linkOutput = messages.LINK_WITH_PROTOCOL_VISITED % link_uri_info[0]
                 if not text:
                     # If there's no text for the link, expose part of the
                     # URI to the user.
@@ -905,55 +927,16 @@ class SpeechGenerator(generator.Generator):
         if _settingsManager.getSetting('onlySpeakDisplayedText'):
             return []
 
-        result = []
-        acss = self.voice(SYSTEM)
         if _settingsManager.getSetting('speechVerbosityLevel') \
-                == settings.VERBOSITY_LEVEL_VERBOSE:
-            if obj.getRole() == pyatspi.ROLE_TABLE_CELL:
-                cell = obj
-            else:
-                cell = self._script.utilities.ancestorWithRole(
-                    obj, [pyatspi.ROLE_TABLE_CELL], [pyatspi.ROLE_FRAME])
-            try:
-                table = cell.parent.queryTable()
-            except:
-                pass
-            else:
-                index = self._script.utilities.cellIndex(cell)
-                row = table.getRowAtIndex(index)
-                col = table.getColumnAtIndex(index)
-                if row + 1 == table.nRows and col + 1 == table.nColumns:
-                    result.append(messages.TABLE_END)
-        if result:
-            result.extend(acss)
-        return result
+           != settings.VERBOSITY_LEVEL_VERBOSE:
+            return []
 
-    #####################################################################
-    #                                                                   #
-    # Terminal information                                              #
-    #                                                                   #
-    #####################################################################
+        if self._script.utilities.isLastCell(obj):
+            result = [messages.TABLE_END]
+            result.extend(self.voice(SYSTEM))
+            return result
 
-    def _generateTerminal(self, obj, **args):
-        """Returns an array of strings (and possibly voice and audio
-        specifications) used especially for handling terminal objects.
-        This either is the name of the frame the terminal is in or the
-        displayed label of the terminal.  [[[WDW - it might be nice
-        to return an empty array if this is not a terminal.]]]
-        """
-        result = []
-        acss = self.voice(DEFAULT)
-        title = None
-        frame = self._script.utilities.ancestorWithRole(
-            obj, [pyatspi.ROLE_FRAME], [])
-        if frame:
-            title = frame.name
-        if not title:
-            title = self._script.utilities.displayedLabel(obj)
-        result.append(title)
-        if result:
-            result.extend(acss)
-        return result
+        return []
 
     #####################################################################
     #                                                                   #
@@ -1233,53 +1216,24 @@ class SpeechGenerator(generator.Generator):
         result.extend(self._getACSS(obj, result[0]))
         return result
 
-    def generateTextIndentation(self, obj, **args):
-        return self._generateTextIndentation(obj, **args)
-
     def _generateTextIndentation(self, obj, **args):
         """Speaks a summary of the number of spaces and/or tabs at the
         beginning of the given line.
 
         Arguments:
         - obj: the text object.
-        - line: the string to check for spaces and tabs.
         """
-        if _settingsManager.getSetting('onlySpeakDisplayedText'):
-            return []
 
-        acss = self.voice(SYSTEM)
         if not _settingsManager.getSetting('enableSpeechIndentation'):
             return []
-        line =  args.get('alreadyFocused', "")
-        if not line:
-            [line, caretOffset, startOffset] = \
-              self._script.getTextLineAtCaret(obj)
 
-        # For the purpose of speaking the text indentation, replace
-        # occurances the non breaking space character with spaces.
-        line = line.replace("\u00a0", " ")
-        spaceCount = 0
-        tabCount = 0
-        utterance = ""
-        offset = 0
-        while True:
-            while (offset < len(line)) and line[offset] == ' ':
-                spaceCount += 1
-                offset += 1
-            if spaceCount:
-                utterance += "%s " % messages.spacesCount(spaceCount)
-            while (offset < len(line)) and line[offset] == '\t':
-                tabCount += 1
-                offset += 1
-            if tabCount:
-                utterance += "%s " % messages.tabsCount(tabCount)
-            if not (spaceCount  or tabCount):
-                break
-            spaceCount  = tabCount = 0
+        line, caretOffset, startOffset = self._script.getTextLineAtCaret(obj)
+        description = self._script.utilities.indentationDescription(line)
+        if not description:
+            return []
 
-        result = [utterance]
-        if result and result[0]:
-            result.extend(acss)
+        result = [description]
+        result.extend(self.voice(SYSTEM))
         return result
 
     #####################################################################
@@ -1324,21 +1278,13 @@ class SpeechGenerator(generator.Generator):
         if _settingsManager.getSetting('onlySpeakDisplayedText'):
             return []
 
-        result = []
-        acss = self.voice(SYSTEM)
-        try:
-            value = obj.queryValue()
-        except NotImplementedError:
-            pass
-        else:
-            percentValue = \
-                (value.currentValue
-                 / (value.maximumValue - value.minimumValue)) \
-                * 100.0
-            result.append(messages.percentage(percentValue))
-        if result:
-            result.extend(acss)
-        return result
+        percentValue = self._script.utilities.getValueAsPercent(obj)
+        if percentValue is not None:
+            result = [messages.percentage(percentValue)]
+            result.extend(self.voice(SYSTEM))
+            return result
+
+        return []
 
     #####################################################################
     #                                                                   #
@@ -1452,13 +1398,14 @@ class SpeechGenerator(generator.Generator):
         if role not in [pyatspi.ROLE_LIST, pyatspi.ROLE_LIST_BOX]:
             return result
 
-        items = self._script.utilities.selectedChildren(obj)
-        items = items or [self._script.utilities.focusedChild(obj)]
+        if 'Selection' in pyatspi.listInterfaces(obj):
+            items = self._script.utilities.selectedChildren(obj)
+        else:
+            items = [self._script.utilities.focusedChild(obj)]
         if not (items and items[0]):
             return result
 
-        items = list(map(self._generateName, items))
-        for item in items:
+        for item in map(self._generateName, items):
             result.extend(item)
 
         return result
@@ -1550,21 +1497,35 @@ class SpeechGenerator(generator.Generator):
         """
         result = []
         priorObj = args.get('priorObj', None)
+        if priorObj and priorObj.getRole() == pyatspi.ROLE_TOOL_TIP:
+            return []
+
         commonAncestor = self._script.utilities.commonAncestor(priorObj, obj)
         try:
             role = commonAncestor.getRole()
+            name = commonAncestor.name
         except:
-            pass
+            role = None
+            name = None
         else:
             if role == pyatspi.ROLE_COMBO_BOX:
                 return []
 
+        def _isCommonAncestor(x):
+            if not (name and role):
+                return False
+            return x and x.getRole() == role and x.name == name
+
         skipRoles = args.get('skipRoles', [])
+        skipRoles.append(pyatspi.ROLE_TREE_ITEM)
         stopAtRoles = args.get('stopAtRoles', [])
         stopAtRoles.append(pyatspi.ROLE_APPLICATION)
         if obj != commonAncestor:
             parent = obj.parent
             while parent and not parent in [commonAncestor, parent.parent]:
+                if _isCommonAncestor(parent):
+                    break
+
                 parentRole = parent.getRole()
                 if parentRole in stopAtRoles:
                     break
@@ -1708,6 +1669,46 @@ class SpeechGenerator(generator.Generator):
         result.extend(acss)
         return result
 
+    def _generateProgressBarIndex(self, obj, **args):
+        if not args.get('isProgressBarUpdate') \
+           or not self._shouldPresentProgressBarUpdate(obj, **args):
+            return []
+
+        result = []
+        acc, updateTime, updateValue = self._getMostRecentProgressBarUpdate()
+        if acc != obj:
+            number, count = self.getProgressBarNumberAndCount(obj)
+            result = [messages.PROGRESS_BAR_NUMBER % (number)]
+            result.extend(self.voice(SYSTEM))
+
+        return result
+
+    def _generateProgressBarValue(self, obj, **args):
+        if args.get('isProgressBarUpdate') \
+           and not self._shouldPresentProgressBarUpdate(obj, **args):
+            return []
+
+        result = []
+        percent = self._script.utilities.getValueAsPercent(obj)
+        if percent is not None:
+            result.append(messages.percentage(percent))
+            result.extend(self.voice(SYSTEM))
+
+        return result
+
+    def _getProgressBarUpdateInterval(self):
+        interval = _settingsManager.getSetting('progressBarSpeechInterval')
+        if interval is None:
+            interval = super()._getProgressBarUpdateInterval()
+
+        return int(interval)
+
+    def _shouldPresentProgressBarUpdate(self, obj, **args):
+        if not _settingsManager.getSetting('speakProgressBarUpdates'):
+            return False
+
+        return super()._shouldPresentProgressBarUpdate(obj, **args)
+
     def _generateDefaultButton(self, obj, **args):
         """Returns an array of strings (and possibly voice and audio
         specifications) that represent the default button in a dialog.
@@ -1719,6 +1720,8 @@ class SpeechGenerator(generator.Generator):
             name = self._generateName(button)
             if name:
                 result.append(messages.DEFAULT_BUTTON_IS % name[0])
+                result.extend(self.voice(SYSTEM))
+
         return result
 
     def generateDefaultButton(self, obj, **args):
@@ -1779,6 +1782,22 @@ class SpeechGenerator(generator.Generator):
             result.append(dialogs)
         return result
 
+    def _generateListBoxItemWidgets(self, obj, **args):
+        widgetRoles = [pyatspi.ROLE_CHECK_BOX,
+                       pyatspi.ROLE_COMBO_BOX,
+                       pyatspi.ROLE_PUSH_BUTTON,
+                       pyatspi.ROLE_RADIO_BUTTON,
+                       pyatspi.ROLE_SLIDER,
+                       pyatspi.ROLE_TOGGLE_BUTTON]
+        isWidget = lambda x: x and x.getRole() in widgetRoles
+        result = []
+        if obj.parent and obj.parent.getRole() == pyatspi.ROLE_LIST_BOX:
+            widgets = pyatspi.findAllDescendants(obj, isWidget)
+            for widget in widgets:
+                result.append(self.generate(widget, includeContext=False))
+
+        return result
+
     #####################################################################
     #                                                                   #
     # Keyboard shortcut information                                     #
@@ -2213,31 +2232,6 @@ class SpeechGenerator(generator.Generator):
     #                                                                   #
     #####################################################################
 
-    def _getAlternativeRole(self, obj, **args):
-        if self._script.utilities.isMath(obj):
-            if self._script.utilities.isMathFraction(obj):
-                return 'ROLE_MATH_FRACTION'
-            if self._script.utilities.isMathRoot(obj):
-                return 'ROLE_MATH_ROOT'
-            if self._script.utilities.isMathSubOrSuperScript(obj):
-                return 'ROLE_MATH_SCRIPT_SUBSUPER'
-            if self._script.utilities.isMathUnderOrOverScript(obj):
-                return 'ROLE_MATH_SCRIPT_UNDEROVER'
-            if self._script.utilities.isMathMultiScript(obj):
-                return 'ROLE_MATH_MULTISCRIPT'
-            if self._script.utilities.isMathEnclose(obj):
-                return 'ROLE_MATH_ENCLOSED'
-            if self._script.utilities.isMathFenced(obj):
-                return 'ROLE_MATH_FENCED'
-            if self._script.utilities.isMathTable(obj):
-                return 'ROLE_MATH_TABLE'
-            if self._script.utilities.isMathTableRow(obj):
-                return 'ROLE_MATH_TABLE_ROW'
-        if self._script.utilities.isStatic(obj):
-            return 'ROLE_STATIC'
-
-        return args.get('role', obj.getRole())
-
     def _generatePause(self, obj, **args):
         if not _settingsManager.getSetting('enablePauseBreaks') \
            or args.get('eliminatePauses', False):
@@ -2256,8 +2250,15 @@ class SpeechGenerator(generator.Generator):
 
         voicename = voiceType.get(key) or voiceType.get(DEFAULT)
         voices = _settingsManager.getSetting('voices')
-        rv = voices.get(voicename)
-        if not rv or rv.get('established') == False:
-            rv = voices.get(voiceType.get(DEFAULT))
+        voice = acss.ACSS(voices.get(voiceType.get(DEFAULT)))
 
-        return [acss.ACSS(rv)]
+        if key in [None, DEFAULT]:
+            string = args.get('string', '')
+            if isinstance(string, str) and string.isupper():
+                voice.update(voices.get(voiceType.get(UPPERCASE)))
+        else:
+            override = voices.get(voicename)
+            if override and override.get('established', True):
+                voice.update(override)
+
+        return [voice]
